@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2023 Gentoo Foundation
+ * Copyright 2005-2025 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2008 Ned Ludd        - <solar@gentoo.org>
@@ -15,6 +15,11 @@
 #include <ctype.h>
 #include <xalloc.h>
 
+#if defined(ENABLE_GPKG) || defined(ENABLE_GTREE)
+# include <archive.h>
+# include <archive_entry.h>
+#endif
+
 #include "atom.h"
 #include "eat_file.h"
 #include "hash.h"
@@ -24,11 +29,8 @@
 #include "tree.h"
 #include "xpak.h"
 
-#include <ctype.h>
-#include <xalloc.h>
 
 static int tree_pkg_compar(const void *l, const void *r);
-static tree_pkg_ctx *tree_next_pkg_int(tree_cat_ctx *cat_ctx);
 static void tree_close_meta(tree_pkg_meta *cache);
 
 static tree_ctx *
@@ -77,50 +79,50 @@ tree_open_int(const char *sroot, const char *tdir, bool quiet)
 
 static const char portcachedir_pms[] = "metadata/cache";
 static const char portcachedir_md5[] = "metadata/md5-cache";
+static const char portcachedir_gt1[] = "metadata/repo.gtree.tar";
 static const char portrepo_name[]    = "profiles/repo_name";
 tree_ctx *
 tree_open(const char *sroot, const char *portdir)
 {
 	tree_ctx *ret;
 	char buf[_Q_PATH_MAX];
-	char *repo = NULL;
-	size_t repolen = 0;
 
-	snprintf(buf, sizeof(buf), "%s%s/%s", sroot, portdir, portrepo_name);
-	if (eat_file(buf, &repo, &repolen)) {
-		(void)rmspace(repo);
-	} else {
-		repo = NULL;  /* ignore missing repo file */
+	ret = tree_open_ebuild(sroot, portdir);
+	if (ret == NULL)
+	{
+		warnf("could not open repository at %s (under root %s)",
+			  portdir, sroot);
+
+		return NULL;
 	}
 
+	/* look for cache trees to speed things up */
+#ifdef ENABLE_GTREE
+	snprintf(buf, sizeof(buf), "%s/%s", portdir, portcachedir_gt1);
+	ret->subtree = tree_open_gtree_int(sroot, buf, true);
+	if (ret->subtree != NULL) {
+		ret->subtree->treetype = TREE_METADATA_GTREE;
+		return ret;
+	}
+#endif
+
 	snprintf(buf, sizeof(buf), "%s/%s", portdir, portcachedir_md5);
-	ret = tree_open_int(sroot, buf, true);
-	if (ret != NULL) {
-		ret->cachetype = CACHE_METADATA_MD5;
-		ret->repo = repo;
+	ret->subtree = tree_open_int(sroot, buf, true);
+	if (ret->subtree != NULL) {
+		ret->subtree->treetype = TREE_METADATA_MD5;
+		ret->subtree->cache.categories = create_set();
 		return ret;
 	}
 
 	snprintf(buf, sizeof(buf), "%s/%s", portdir, portcachedir_pms);
-	ret = tree_open_int(sroot, buf, true);
-	if (ret != NULL) {
-		ret->cachetype = CACHE_METADATA_PMS;
-		ret->repo = repo;
+	ret->subtree = tree_open_int(sroot, buf, true);
+	if (ret->subtree != NULL) {
+		ret->subtree->treetype = TREE_METADATA_PMS;
+		ret->subtree->cache.categories = create_set();
 		return ret;
 	}
 
-	ret = tree_open_int(sroot, portdir, true);
-	if (ret != NULL) {
-		ret->cachetype = CACHE_EBUILD;
-		ret->repo = repo;
-		return ret;
-	}
-
-	if (repo != NULL)
-		free(repo);
-	warnf("could not open repository at %s (under root %s)", portdir, sroot);
-
-	return NULL;
+	return ret;
 }
 
 tree_ctx *
@@ -128,7 +130,7 @@ tree_open_vdb(const char *sroot, const char *svdb)
 {
 	tree_ctx *ret = tree_open_int(sroot, svdb, false);
 	if (ret != NULL)
-		ret->cachetype = CACHE_VDB;
+		ret->treetype = TREE_VDB;
 	return ret;
 }
 
@@ -146,7 +148,7 @@ tree_open_ebuild(const char *sroot, const char *portdir)
 			(void)rmspace(repo);
 			ret->repo = repo;
 		}
-		ret->cachetype = CACHE_EBUILD;
+		ret->treetype = TREE_EBUILD;
 	}
 	return ret;
 }
@@ -159,11 +161,11 @@ tree_open_binpkg(const char *sroot, const char *spkg)
 	int fd;
 
 	if (ret != NULL) {
-		ret->cachetype = CACHE_BINPKGS;
+		ret->treetype = TREE_BINPKGS;
 
 		fd = openat(ret->tree_fd, binpkg_packages, O_RDONLY | O_CLOEXEC);
 		if (eat_file_fd(fd, &ret->cache.store, &ret->cache.storesize)) {
-			ret->cachetype = CACHE_PACKAGES;
+			ret->treetype = TREE_PACKAGES;
 		} else if (ret->cache.store != NULL) {
 			free(ret->cache.store);
 			ret->cache.store = NULL;
@@ -174,6 +176,70 @@ tree_open_binpkg(const char *sroot, const char *spkg)
 	return ret;
 }
 
+#ifdef ENABLE_GTREE
+tree_ctx *
+tree_open_gtree_int(const char *sroot, const char *tar, bool quiet)
+{
+	tree_ctx             *ctx = xzalloc(sizeof(*ctx));
+	struct archive       *gt;
+	struct archive_entry *entry;
+
+	ctx->portroot_fd = open(sroot, O_RDONLY | O_CLOEXEC | O_PATH);
+	if (ctx->portroot_fd == -1) {
+		if (!quiet)
+			warnp("could not open root: %s", sroot);
+		free(ctx);
+		return NULL;
+	}
+
+	/* Skip the leading slash */
+	tar++;
+	if (*tar == '\0')
+		tar = ".";
+	ctx->tree_fd = openat(ctx->portroot_fd, tar, O_RDONLY | O_CLOEXEC);
+	if (ctx->tree_fd == -1) {
+		if (!quiet)
+			warnp("could not open tree: %s (in root %s)", tar, sroot);
+		close(ctx->portroot_fd);
+		free(ctx);
+		return NULL;
+	}
+
+	if (sroot[0] == '/' && sroot[1] == '\0')
+		sroot = "";
+	snprintf(ctx->path, sizeof(ctx->path), "%s/%s", sroot, tar);
+
+	gt = archive_read_new();
+	archive_read_support_format_all(gt);
+	if (archive_read_open_fd(gt, ctx->tree_fd, BUFSIZ) != ARCHIVE_OK ||
+		archive_read_next_header(gt, &entry) != ARCHIVE_OK)
+	{
+		if (!quiet)
+			warnp("could not open tree: %s (in root %s): %s",
+				  tar, sroot, archive_error_string(gt));
+		archive_read_free(gt);
+		close(ctx->portroot_fd);
+		close(ctx->tree_fd);
+		free(ctx);
+		return NULL;
+	}
+	if (strcmp(archive_entry_pathname(entry), "gtree-1") != 0) {
+		if (!quiet)
+			warnp("could not open tree: %s (in root %s): not a gtree container",
+				  tar, sroot);
+		archive_read_free(gt);
+		close(ctx->portroot_fd);
+		close(ctx->tree_fd);
+		free(ctx);
+		return NULL;
+	}
+	/* defer repo for now */
+	archive_read_free(gt);
+
+	return ctx;
+}
+#endif
+
 void
 tree_close(tree_ctx *ctx)
 {
@@ -197,9 +263,10 @@ tree_close(tree_ctx *ctx)
 	if (ctx->cache.store != NULL)
 		free(ctx->cache.store);
 
-	closedir(ctx->dir);
-	/* closedir() above does this for us: */
-	/* close(ctx->tree_fd); */
+	if (ctx->dir != NULL)
+		closedir(ctx->dir);
+	else if (ctx->tree_fd >= 0)
+		close(ctx->tree_fd); /* closedir() above does this for us */
 	close(ctx->portroot_fd);
 	if (ctx->do_sort)
 		scandir_free(ctx->cat_de, ctx->cat_cnt);
@@ -209,6 +276,8 @@ tree_close(tree_ctx *ctx)
 		free(ctx->pkgs);
 	if (ctx->ebuilddir_ctx != NULL)
 		free(ctx->ebuilddir_ctx);
+	if (ctx->subtree != NULL)
+		tree_close(ctx->subtree);
 	free(ctx);
 }
 
@@ -279,7 +348,7 @@ tree_open_cat(tree_ctx *ctx, const char *name)
 		return NULL;
 	}
 
-	cat_ctx = xmalloc(sizeof(*cat_ctx));
+	cat_ctx = xzalloc(sizeof(*cat_ctx));
 	cat_ctx->name = name;
 	cat_ctx->fd = fd;
 	cat_ctx->dir = dir;
@@ -327,8 +396,14 @@ tree_next_cat(tree_ctx *ctx)
 							(int (*)(const void *, const void *))alphasort);
 				free(cats);
 			} else {
-				ctx->cat_cnt = scandirat(ctx->tree_fd,
-						".", &ctx->cat_de, tree_filter_cat, alphasort);
+				int sdret = scandirat(ctx->tree_fd, ".",
+									  &ctx->cat_de,
+									  tree_filter_cat, alphasort);
+				/* CID 548426 */
+				if (sdret < 0)
+					ctx->cat_cnt = 0;
+				else
+					ctx->cat_cnt = (size_t)sdret;
 			}
 			ctx->cat_cur = 0;
 		}
@@ -389,6 +464,7 @@ tree_close_cat(tree_cat_ctx *cat_ctx)
 	/* close(ctx->fd); */
 	if (cat_ctx->ctx->do_sort)
 		free(cat_ctx->pkg_ctxs);
+
 	free(cat_ctx);
 }
 
@@ -421,17 +497,50 @@ tree_filter_pkg(const struct dirent *de)
 		}
 	}
 
+	if (i > 0 &&
+		(strcmp(de->d_name, "Manifest") == 0 ||
+		 strcmp(de->d_name, "metadata.xml") == 0))
+		i = 0;
+
 	return i;
 }
 
 tree_pkg_ctx *
 tree_open_pkg(tree_cat_ctx *cat_ctx, const char *name)
 {
-	tree_pkg_ctx *pkg_ctx = xzalloc(sizeof(*pkg_ctx));
-	pkg_ctx->name = name;
-	pkg_ctx->repo = cat_ctx->ctx->repo;
-	pkg_ctx->fd = -1;
-	pkg_ctx->cat_ctx = cat_ctx;
+	tree_pkg_ctx *pkg_ctx;
+	bool          isgpkg = false;
+
+	if (cat_ctx->ctx->treetype == TREE_EBUILD &&
+		cat_ctx->ctx->ebuilddir_cat_ctx == cat_ctx)
+	{
+		size_t len = strlen(name);
+		char  *p   = (char *)&name[len - (sizeof(".ebuild") - 1)];
+		if (len <= sizeof(".ebuild") - 1 ||
+			memcmp(p, ".ebuild", sizeof(".ebuild") - 1) != 0)
+			return NULL;  /* invalid, must be some random other file */
+		*p = '\0';
+	} else if (cat_ctx->ctx->treetype == TREE_BINPKGS) {
+		size_t len = strlen(name);
+		char  *p   = (char *)&name[len - (sizeof(".gpkg.tar") - 1)];
+		if (len > sizeof(".gpkg.tar") - 1 &&
+			memcmp(p, ".gpkg.tar", sizeof(".gpkg.tar") - 1) == 0)
+			isgpkg = true;
+
+		if (!isgpkg &&
+			(len <= sizeof(".tbz2") - 1 ||
+			 (p = (char *)&name[len - (sizeof(".tbz2") - 1)]) == NULL ||
+			 memcmp(p, ".tbz2", sizeof(".tbz2") - 1) != 0))
+			return NULL;  /* invalid, like above */
+		*p = '\0';
+	}
+
+	pkg_ctx                = xzalloc(sizeof(*pkg_ctx));
+	pkg_ctx->name          = name;
+	pkg_ctx->repo          = cat_ctx->ctx->repo;
+	pkg_ctx->fd            = -1;
+	pkg_ctx->cat_ctx       = cat_ctx;
+	pkg_ctx->binpkg_isgpkg = isgpkg;
 
 	/* see if this pkg matches the query, here we can finally check
 	 * version conditions like >=, etc. */
@@ -496,7 +605,7 @@ tree_next_pkg_int(tree_cat_ctx *cat_ctx)
 				pkg_ctx = cat_ctx->pkg_ctxs[cat_ctx->pkg_cnt++] =
 					tree_open_pkg(cat_ctx, name);
 				if (pkg_ctx == NULL) {
-					/* name was freed by tree_close_pkg on fail */
+					free(name);
 					cat_ctx->pkg_cnt--;
 				}
 			}
@@ -541,12 +650,7 @@ tree_next_pkg(tree_cat_ctx *cat_ctx)
 	tree_ctx *ctx = cat_ctx->ctx;
 	tree_pkg_ctx *ret = NULL;
 
-	if (ctx->do_sort && cat_ctx->pkg_ctxs != NULL) {
-		/* bypass to use the cache if it exists */
-		ret = tree_next_pkg_int(cat_ctx);
-	} else if (ctx->cachetype == CACHE_EBUILD) {
-		char *p;
-
+	if (ctx->treetype == TREE_EBUILD) {
 		/* serve *.ebuild files each as separate pkg_ctx with name set
 		 * to CAT/P like in VDB and metadata */
 		do {
@@ -554,17 +658,20 @@ tree_next_pkg(tree_cat_ctx *cat_ctx)
 				tree_ctx *pkgdir = ctx->ebuilddir_ctx;
 
 				if (pkgdir == NULL)
-					pkgdir = ctx->ebuilddir_ctx = xzalloc(sizeof(tree_ctx));
+					pkgdir = ctx->ebuilddir_ctx = xmalloc(sizeof(*pkgdir));
 
+				ctx->ebuilddir_cat_ctx = NULL;
 				ctx->ebuilddir_pkg_ctx = tree_next_pkg_int(cat_ctx);
 				if (ctx->ebuilddir_pkg_ctx == NULL)
 					return NULL;
 
+				memset(pkgdir, 0, sizeof(*pkgdir));
 				pkgdir->portroot_fd = -1;
 				pkgdir->tree_fd = cat_ctx->fd;
 				pkgdir->do_sort = ctx->do_sort;
 				pkgdir->repo = ctx->repo;
-				pkgdir->cachetype = ctx->cachetype;
+				pkgdir->treetype = ctx->treetype;
+				pkgdir->subtree = ctx->subtree;
 
 				ctx->ebuilddir_cat_ctx =
 					tree_open_cat(pkgdir, ctx->ebuilddir_pkg_ctx->name);
@@ -572,37 +679,24 @@ tree_next_pkg(tree_cat_ctx *cat_ctx)
 				/* opening might fail if what we found wasn't a
 				 * directory or something */
 				if (ctx->ebuilddir_cat_ctx == NULL) {
+					tree_close_pkg(ctx->ebuilddir_pkg_ctx);
 					ctx->ebuilddir_pkg_ctx = NULL;
 					continue;
 				}
 
 				/* "zap" the pkg such that it looks like CAT/P */
 				ctx->ebuilddir_cat_ctx->name = cat_ctx->name;
+				ctx->ebuilddir_cat_ctx->ctx = ctx;
 			}
 
 			ret = tree_next_pkg_int(ctx->ebuilddir_cat_ctx);
 			if (ret == NULL) {
 				tree_close_cat(ctx->ebuilddir_cat_ctx);
+				if (!cat_ctx->ctx->do_sort || cat_ctx->pkg_ctxs == NULL)
+					tree_close_pkg(ctx->ebuilddir_pkg_ctx);
 				ctx->ebuilddir_pkg_ctx = NULL;
-			} else {
-				if ((p = strstr(ret->name, ".ebuild")) == NULL) {
-					tree_close_pkg(ret);
-					ret = NULL;
-				} else {
-					*p = '\0';
-				}
 			}
 		} while (ret == NULL);
-	} else if (ctx->cachetype == CACHE_BINPKGS) {
-		char *p = NULL;
-		ret = NULL;
-		do {
-			if (ret != NULL)
-				tree_close_pkg(ret);
-			ret = tree_next_pkg_int(cat_ctx);
-		} while (ret != NULL && (p = strstr(ret->name, ".tbz2")) == NULL);
-		if (p != NULL)
-			*p = '\0';
 	} else {
 		ret = tree_next_pkg_int(cat_ctx);
 	}
@@ -916,13 +1010,18 @@ tree_read_file_ebuild(tree_pkg_ctx *pkg_ctx)
 								continue;
 							}
 						} else {
-							/* implement line continuation (\ before newline) */
-							if (esc && (*p == '\n' || *p == '\r'))
+							/* stash everything on a single line like
+							 * VDB and md5-cache do */
+							if (*p == '\n' || *p == '\r')
 								*p = ' ';
 							esc = false;
 						}
 
-						*w++ = *p++;
+						/* collapse sequences of spaces */
+						if (*w != ' ' || *p != ' ')
+							*w++ = *p++;
+						else
+							p++;
 					}
 					if (*p == *q && esc) {
 						/* escaped, move along */
@@ -1036,17 +1135,89 @@ static tree_pkg_meta *
 tree_read_file_binpkg(tree_pkg_ctx *pkg_ctx)
 {
 	tree_pkg_meta *m = xzalloc(sizeof(tree_pkg_meta));
-	int newfd = dup(pkg_ctx->fd);
+	int newfd = -1;
 
-	xpak_process_fd(pkg_ctx->fd, true, m, tree_read_file_binpkg_xpak_cb);
-	pkg_ctx->fd = -1;  /* closed by xpak_process_fd */
+	if (pkg_ctx->binpkg_isgpkg) {
+#ifdef ENABLE_GPKG
+		struct archive       *a     = archive_read_new();
+		struct archive_entry *entry;
+		size_t                len   = 0;
+		char                 *buf   = NULL;
+
+		archive_read_support_format_all(a);
+		archive_read_support_filter_all(a);
+
+		if (archive_read_open_fd(a, pkg_ctx->fd, BUFSIZ) != ARCHIVE_OK)
+			return NULL;
+		while (archive_read_next_header(a, &entry) == ARCHIVE_OK) {
+			const char *pathname = archive_entry_pathname(entry);
+			const char *fname    = strchr(pathname, '/');
+			if (fname == NULL)
+				continue;
+			fname++;
+			if (strncmp(fname, "metadata.tar",
+						sizeof("metadata.tar") - 1) == 0)
+			{
+				/* read this nested tar, it contains the VDB entries
+				 * otherwise stored in xpak */
+				len = archive_entry_size(entry);
+				buf = xmalloc(len);
+				archive_read_data(a, buf, len);
+				break;
+			}
+		}
+		archive_read_free(a);
+		newfd = pkg_ctx->fd;
+		pkg_ctx->fd = -1;  /* will be closed by hash_multiple_fd */
+
+		if (buf != NULL)
+		{
+			char  *data      = NULL;
+			size_t data_size = 0;
+			size_t data_len  = 0;
+
+			a = archive_read_new();
+			archive_read_support_format_all(a);
+			archive_read_support_filter_all(a);
+			archive_read_open_memory(a, buf, len);
+
+			while (archive_read_next_header(a, &entry) == ARCHIVE_OK) {
+				const char *pathname = archive_entry_pathname(entry);
+				char       *fname    = strchr(pathname, '/');
+				if (fname == NULL)
+					continue;
+				fname++;
+
+				data_len = archive_entry_size(entry);
+				if (data_len > data_size) {
+					data_size = data_len;
+					data      = xrealloc(data, data_size);
+				}
+				if (archive_read_data(a, data, data_len) < 0)
+					continue;
+				tree_read_file_binpkg_xpak_cb(m,
+											  fname, (int)strlen(fname),
+											  0, data_len, data);
+			}
+			archive_read_free(a);
+			free(buf);
+			free(data);
+		}
+#else
+		return NULL;
+#endif
+	} else {
+		newfd = dup(pkg_ctx->fd);
+		xpak_process_fd(pkg_ctx->fd, true, m, tree_read_file_binpkg_xpak_cb);
+		pkg_ctx->fd = -1;  /* closed by xpak_process_fd */
+	}
 
 	/* fill in some properties which are not available, but would be in
 	 * Packages, and used to verify the package ... this is somewhat
 	 * fake, but allows to transparantly use a dir of binpkgs */
 	if (newfd != -1) {
 		size_t fsize;
-		size_t needlen = 40 + 1 + 19 + 1;
+		size_t needlen = SHA1_DIGEST_LENGTH + 1 + 19 + 1;
 		size_t pos = 0;
 		size_t len = 0;
 
@@ -1062,10 +1233,10 @@ tree_read_file_binpkg(tree_pkg_ctx *pkg_ctx)
 		}
 
 		m->Q_SHA1 = m->storage->ptr + pos;
-		m->Q_SIZE = m->Q_SHA1 + 40 + 1;
+		m->Q_SIZE = m->Q_SHA1 + SHA1_DIGEST_LENGTH + 1;
 		m->storage->pos += needlen;
 
-		lseek(newfd, 0, SEEK_SET);  /* reposition at the whole file */
+		lseek(newfd, 0, SEEK_SET);  /* reposition at the start of file */
 		if (hash_multiple_file_fd(newfd, NULL, m->Q_SHA1, NULL, NULL,
 				NULL, &fsize, HASH_SHA1) == 0)
 			snprintf(m->Q_SIZE, 19 + 1, "%zu", fsize);
@@ -1074,6 +1245,30 @@ tree_read_file_binpkg(tree_pkg_ctx *pkg_ctx)
 	return m;
 }
 
+static int
+tree_pkg_read_openfd_int(tree_pkg_ctx *pkg_ctx)
+{
+	tree_ctx *ctx = pkg_ctx->cat_ctx->ctx;
+
+	if (pkg_ctx->fd == -1) {
+		if (ctx->treetype == TREE_EBUILD ||
+			ctx->treetype == TREE_BINPKGS)
+		{
+			char buf[_Q_PATH_MAX];
+			snprintf(buf, sizeof(buf), "%s.%s", pkg_ctx->name,
+					 ctx->treetype == TREE_EBUILD ? "ebuild" :
+					 pkg_ctx->binpkg_isgpkg ? "gpkg.tar" : "tbz2");
+			pkg_ctx->fd = openat(pkg_ctx->cat_ctx->fd, buf,
+								 O_RDONLY | O_CLOEXEC);
+		} else {
+			pkg_ctx->fd = openat(pkg_ctx->cat_ctx->fd, pkg_ctx->name,
+								 O_RDONLY | O_CLOEXEC);
+		}
+	}
+
+	return pkg_ctx->fd;
+}
+
 static tree_pkg_meta *
 tree_pkg_read(tree_pkg_ctx *pkg_ctx)
 {
@@ -1083,40 +1278,88 @@ tree_pkg_read(tree_pkg_ctx *pkg_ctx)
 	if (pkg_ctx->meta != NULL)
 		return pkg_ctx->meta;
 
-	if (pkg_ctx->fd == -1) {
-		if (ctx->cachetype == CACHE_EBUILD || ctx->cachetype == CACHE_BINPKGS) {
-			char *p = (char *)pkg_ctx->name;
-			p += strlen(p);
-			*p = '.';
-			pkg_ctx->fd = openat(pkg_ctx->cat_ctx->fd, pkg_ctx->name,
-					O_RDONLY | O_CLOEXEC);
-			*p = '\0';
-		} else {
-			pkg_ctx->fd = openat(pkg_ctx->cat_ctx->fd, pkg_ctx->name,
-					O_RDONLY | O_CLOEXEC);
-		}
-		if (pkg_ctx->fd == -1)
-			return NULL;
-	}
+	if (tree_pkg_read_openfd_int(pkg_ctx) == -1)
+		return NULL;
 
-	if (ctx->cachetype == CACHE_METADATA_MD5) {
+	if (ctx->treetype == TREE_METADATA_MD5) {
 		ret = tree_read_file_md5(pkg_ctx);
 		/* md5-cache, is sort of documented in egencache man-page
 		 * key-points are that an md5 is provided for the ebuild itself,
 		 * and if it includes eclasses, the md5s for each eclass.  These
 		 * are available as _md5_ and _eclasses_ keys.  The latter uses
 		 * tab-separation of form <eclass-name>\t<md5>\t... */
-	} else if (ctx->cachetype == CACHE_METADATA_PMS) {
+	} else if (ctx->treetype == TREE_METADATA_PMS) {
 		ret = tree_read_file_pms(pkg_ctx);
 		/* PMS implies to do an mtime and existence check (the cache may
 		 * contain extra stuff) but since this form of metadata in fact
 		 * is extinct, because these checks are insufficient and
 		 * impossible on e.g. a git-based tree. */
-	} else if (ctx->cachetype == CACHE_EBUILD) {
-		ret = tree_read_file_ebuild(pkg_ctx);
-	} else if (ctx->cachetype == CACHE_BINPKGS) {
+	} else if (ctx->treetype == TREE_EBUILD) {
+		ret = NULL;
+		if (ctx->subtree != NULL) {
+			tree_cat_ctx  *scat;
+			tree_pkg_ctx  *spkg  = NULL;
+
+			/* the cache takes care of repeated calls here */
+			scat = tree_open_cat(ctx->subtree, pkg_ctx->cat_ctx->name);
+			if (scat != NULL)
+				spkg = tree_open_pkg(scat, pkg_ctx->name);
+
+			if (spkg != NULL) {
+				if (ctx->subtree->treetype == TREE_METADATA_MD5) {
+					/* in this case a cache entry exists, however, it
+					 * may be out of date, for that we need to check the
+					 * md5 hashes with the ebuild/eclass files,
+					 * obviously when the source ebuild doesn't exist,
+					 * we never get here */
+					char   *mdmd5;
+					char    srcmd5[MD5_DIGEST_LENGTH + 1];
+					size_t  flen;
+
+					if (hash_multiple_file_fd(pkg_ctx->fd,
+										  	  srcmd5, NULL, NULL, NULL,
+										  	  NULL, &flen, HASH_MD5) == 0)
+						pkg_ctx->fd = -1;
+
+					mdmd5 = tree_pkg_meta_get(spkg, _md5_);
+					/* TODO: eclass compares */
+
+					/* is this a valid cache? use it! */
+					if (mdmd5 != NULL &&
+						memcmp(mdmd5, srcmd5, MD5_DIGEST_LENGTH) == 0)
+					{
+						ret = tree_pkg_read(spkg);
+					}
+				} else if (ctx->subtree->treetype == TREE_METADATA_PMS) {
+					struct stat ebld;
+					struct stat pmsc;
+
+					if (fstat(pkg_ctx->fd, &ebld) != 0 ||
+						fstat(tree_pkg_read_openfd_int(spkg), &pmsc) != 0 ||
+						ebld.st_mtime > pmsc.st_mtime ||
+						(ebld.st_mtime == pmsc.st_mtime &&
+						 ebld.st_mtim.tv_nsec > pmsc.st_mtim.tv_nsec))
+					{
+						/* fail or ebuild is newer, so ignore */
+					} else {
+						ret = tree_pkg_read(spkg);
+					}
+				}
+				if (ret != NULL) {
+					/* transplant meta to pkg, so we can free spkg */
+					pkg_ctx->meta = spkg->meta;
+					spkg->meta    = NULL;
+				}
+				tree_close_pkg(spkg);
+			}
+		}
+		if (ret == NULL) {
+			(void)tree_pkg_read_openfd_int(pkg_ctx);  /* re-open if fallback */
+			ret = tree_read_file_ebuild(pkg_ctx);
+		}
+	} else if (ctx->treetype == TREE_BINPKGS) {
 		ret = tree_read_file_binpkg(pkg_ctx);
-	} else if (ctx->cachetype == CACHE_PACKAGES) {
+	} else if (ctx->treetype == TREE_PACKAGES) {
 		ret = (tree_pkg_meta *)pkg_ctx->cat_ctx->ctx->pkgs;
 	}
 
@@ -1182,7 +1425,7 @@ tree_pkg_meta_get_int(tree_pkg_ctx *pkg_ctx, size_t offset, const char *keyn)
 	 * key, the tree_pkg_meta_get macro as called by the user takes care
 	 * of offset and keyn pointing to the same thing */
 
-	if (ctx->cachetype == CACHE_VDB) {
+	if (ctx->treetype == TREE_VDB) {
 		if (pkg_ctx->meta == NULL)
 			pkg_ctx->meta = xzalloc(sizeof(tree_pkg_meta));
 
@@ -1245,8 +1488,8 @@ tree_pkg_meta_get_int(tree_pkg_ctx *pkg_ctx, size_t offset, const char *keyn)
 		 * contain everything available (for a semi-good reason though)
 		 * We cannot downgrade the tree execution to BINPKGS, because
 		 * we're running from tree_foreach_packages */
-		if (*key == NULL && ctx->cachetype == CACHE_PACKAGES) {
-			ctx->cachetype = CACHE_BINPKGS;
+		if (*key == NULL && ctx->treetype == TREE_PACKAGES) {
+			ctx->treetype = TREE_BINPKGS;
 			pkg_ctx->fd = -1;
 
 			/* trigger tree_pkg_read to do something */
@@ -1255,7 +1498,7 @@ tree_pkg_meta_get_int(tree_pkg_ctx *pkg_ctx, size_t offset, const char *keyn)
 			pkg_ctx->meta = NULL;
 
 			pkg_ctx->meta = tree_pkg_read(pkg_ctx);
-			ctx->cachetype = CACHE_PACKAGES;
+			ctx->treetype = TREE_PACKAGES;
 			if (pkg_ctx->meta == NULL) {
 				/* hrmffff. */
 				pkg_ctx->fd = -2;
@@ -1287,7 +1530,7 @@ tree_pkg_metadata(tree_pkg_ctx *pkg_ctx)
 	 * tag, but practically speaking we don't care at all, so we can
 	 * just extract everything between <email> and </email> */
 
-	if (ctx->cachetype == CACHE_EBUILD) {
+	if (ctx->treetype == TREE_EBUILD) {
 		fd = openat(pkg_ctx->cat_ctx->fd, "metadata", O_RDONLY | O_CLOEXEC);
 	} else {
 		char buf[_Q_PATH_MAX];
@@ -1453,9 +1696,30 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 				}
 				if (meta.Q_BUILDID != NULL)
 					atom->BUILDID = atoi(meta.Q_BUILDID);
-				pkgnamelen = snprintf(pkgname, sizeof(pkgname),
-						"%s.tbz2", atom->PF);
-				pkgname[pkgnamelen - (sizeof(".tbz2") - 1)] = '\0';
+				pkgnamelen = 0;
+				if (meta.Q_PATH != NULL) {
+					size_t plen = strlen(meta.Q_PATH);
+					if (plen > sizeof(".tbz2") - 1 &&
+						memcmp(meta.Q_PATH + plen - sizeof(".tbz2") - 1,
+							   ".tbz2", sizeof(".tbz2") - 1) == 0)
+					{
+						pkgnamelen = snprintf(pkgname, sizeof(pkgname),
+											  "%s.tbz2", atom->PF);
+						pkgname[pkgnamelen - (sizeof(".tbz2") - 1)] = '\0';
+					}
+					if (plen > sizeof(".gpkg.tar") - 1 &&
+						memcmp(meta.Q_PATH + plen - sizeof(".gpkg.tar") - 1,
+							   ".gpkg.tar", sizeof(".gpkg.tar") - 1) == 0)
+					{
+						pkgnamelen = snprintf(pkgname, sizeof(pkgname),
+											  "%s.gpkg.tar", atom->PF);
+						pkgname[pkgnamelen - (sizeof(".gpkg.tar") - 1)] = '\0';
+					}
+				}
+				if (pkgnamelen == 0) {
+					pkgnamelen = snprintf(pkgname, sizeof(pkgname),
+										  "%s", atom->PF);
+				}
 				pkg.name = pkgname;
 				pkg.slot = meta.Q_SLOT == NULL ? (char *)"0" : meta.Q_SLOT;
 				pkg.repo = ctx->repo;
@@ -1551,6 +1815,227 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 	return ret;
 }
 
+#ifdef ENABLE_GTREE
+struct tree_gtree_cb_ctx {
+	struct archive *archive;
+};
+static la_ssize_t
+tree_gtree_read_cb(struct archive *a, void *cctx, const void **buf)
+{
+	struct tree_gtree_cb_ctx *ctx = cctx;
+	size_t                    size;
+	la_int64_t                offset;  /* unused */
+	int                       ret;
+
+	(void)a;
+
+	ret = archive_read_data_block(ctx->archive, buf, &size, &offset);
+	if (ret == ARCHIVE_EOF)
+		return 0;
+	if (ret != ARCHIVE_OK)
+		return -1;
+	(void)offset;
+	/* at this point I sincerely hope size is not going to be over the
+	 * unsigned variant */
+	return (la_ssize_t)size;
+}
+static int
+tree_gtree_close_cb(struct archive *a, void *cctx)
+{
+	(void)a;
+	(void)cctx;
+
+	/* noop */
+	return ARCHIVE_OK;
+}
+
+static int
+tree_foreach_gtree(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
+{
+	tree_cat_ctx             *cat         = NULL;
+	tree_pkg_ctx              pkg;
+	tree_pkg_meta             meta;
+	depend_atom              *atom        = NULL;
+	struct archive           *outer;
+	struct archive           *inner;
+	struct archive_entry     *entry;
+	struct tree_gtree_cb_ctx  cb_ctx;
+	char                     *p;
+	const depend_atom        *query       = ctx->query_atom;
+	bool                      foundcaches = false;
+	size_t                    len;
+	int                       ret         = 0;
+
+	/* reused for every entry */
+	cat       = xzalloc(sizeof(*cat));
+	cat->ctx  = ctx;
+	cat->name = "";
+
+	/* rewind the outer tar, and re-read it, it's slight overhead that
+	 * we have to re-read it entirely, but I cannot find an API to
+	 * rewind or use an offset to re-read an entry again, and the
+	 * overhead should be small given the outer is uncompressed and it's
+	 * the second entry */
+	lseek(ctx->tree_fd, 0, SEEK_SET);
+	outer = archive_read_new();
+	archive_read_support_format_all(outer);  /* don't see why not */
+	if (archive_read_open_fd(outer, ctx->tree_fd, BUFSIZ) != ARCHIVE_OK) {
+		warn("unable to read gtree container: %s",
+			 archive_error_string(outer));
+		archive_read_free(outer);
+		return 1;
+	}
+
+	while (archive_read_next_header(outer, &entry) == ARCHIVE_OK) {
+		const char *fname = archive_entry_pathname(entry);
+		if (fname == NULL)
+			continue;
+		if (strncmp(fname, "repo.tar", sizeof("repo.tar") - 1) == 0 &&
+			(fname[sizeof("repo.tar") - 1] == '.' ||
+			 fname[sizeof("repo.tar") - 1] == '\0'))
+			break;
+		entry = NULL;
+	}
+	if (entry == NULL) {
+		archive_read_free(outer);
+		return 1;
+	}
+
+	/* use wrapper to read straight from this archive */
+	inner = archive_read_new();
+	archive_read_support_format_all(inner);
+	archive_read_support_filter_all(inner);
+	memset(&cb_ctx, 0, sizeof(cb_ctx));
+	cb_ctx.archive = outer;
+	if (archive_read_open(inner, &cb_ctx, NULL,
+						  tree_gtree_read_cb,
+						  tree_gtree_close_cb) != ARCHIVE_OK)
+		warn("unable to read gtree data %s: %s", archive_entry_pathname(entry),
+			 archive_error_string(inner));
+
+	while (archive_read_next_header(inner, &entry) == ARCHIVE_OK) {
+		const char *fname = archive_entry_pathname(entry);
+
+		if (fname == NULL)
+			continue;
+
+		if (strncmp(fname, "caches/", sizeof("caches/") - 1) == 0) {
+			char *nexttok = NULL;
+
+			foundcaches = true;
+			fname += sizeof("caches/") - 1;
+			atom = atom_explode(fname);
+			if (query != NULL && atom_compare(atom, query) != EQUAL)
+				continue;
+
+			/* ok, we're in business */
+			len = archive_entry_size(entry);
+			if (len > ctx->cache.storesize) {
+				ctx->cache.storesize = len;
+				ctx->cache.store     = xrealloc(ctx->cache.store, len);
+			}
+			archive_read_data(inner, ctx->cache.store, ctx->cache.storesize);
+
+			memset(&meta, 0, sizeof(meta));
+			/* entries are strictly single line, starting with KEY= (no
+			 * whitespace) */
+			for (p = strtok_r(ctx->cache.store, "=", &nexttok);
+				 p != NULL;
+				 p = strtok_r(NULL, "=", &nexttok))
+			{
+				char **ptr = NULL;
+
+				if (1 == 0) {
+					/* dummy case for syntax */
+				}
+#define match2(K,N) \
+				else if (strcmp(p, #N) == 0) \
+					ptr = &meta.Q_##K
+#define match(K)  match2(K,K)
+				match(DEPEND);
+				match(RDEPEND);
+				match(SLOT);
+				match(SRC_URI);
+				match(RESTRICT);
+				match(LICENSE);
+				match(DESCRIPTION);
+				match(KEYWORDS);
+				match(INHERITED);
+				match(IUSE);
+				match(CDEPEND);
+				match(PDEPEND);
+				match(PROVIDE);
+				match(EAPI);
+				match(PROPERTIES);
+				match(BDEPEND);
+				match(IDEPEND);
+				match(DEFINED_PHASES);
+				match(REQUIRED_USE);
+				match(CONTENTS);
+				match(USE);
+				match(EPREFIX);
+				match(PATH);
+				match(BUILDID);
+				match(SIZE);
+				match2(_eclasses_, eclasses);
+#undef match
+#undef match2
+
+				/* always advance to end of line, even when nothing
+				 * matched */
+				p = strtok_r(NULL, "\n", &nexttok);
+				if (p == NULL)
+					break;
+				if (ptr != NULL)
+					*ptr = p;
+			}
+
+			pkg.name    = atom->PF;
+			pkg.slot    = meta.Q_SLOT == NULL ? (char *)"0" : meta.Q_SLOT;
+			pkg.repo    = ctx->repo;
+			pkg.atom    = atom;
+			pkg.cat_ctx = cat;
+			pkg.fd      = -2;  /* intentional, meta has already been read */
+
+			if (strcmp(cat->name, atom->CATEGORY) != 0)
+			{
+				if (cat->pkg_ctxs != NULL) {
+					atom_implode((depend_atom *)cat->pkg_ctxs);
+					cat->pkg_ctxs = NULL;
+				}
+				cat->name     = atom->CATEGORY;
+				cat->pkg_ctxs = (tree_pkg_ctx **)atom;  /* for name */
+				atom          = NULL;  /* avoid double free */
+			}
+
+			/* do call callback with pkg_atom (populate cat and pkg) */
+			ret |= callback(&pkg, priv);
+
+			if (atom != NULL)
+				atom_implode(atom);
+		} else if (foundcaches) {
+			break;  /* stop searching if we processed all cache entries */
+		} else if (ctx->repo == NULL &&
+				   strcmp(fname, "repository") == 0)
+		{
+			/* fill in repo, so it can be used when requested */
+			len = archive_entry_size(entry);
+			ctx->repo = xmalloc(len);
+			archive_read_data(inner, ctx->repo, len);
+		}
+	}
+
+	/* ensure we don't free double */
+	ctx->repo = NULL;
+	ctx->do_sort = false;
+	if (ctx->cache.storesize > 0)
+		ctx->cache.store[0] = '\0';
+	free(cat);
+
+	return ret;
+}
+#endif
+
 int
 tree_foreach_pkg(tree_ctx *ctx, tree_pkg_cb callback, void *priv,
 		bool sort, const depend_atom *query)
@@ -1566,9 +2051,19 @@ tree_foreach_pkg(tree_ctx *ctx, tree_pkg_cb callback, void *priv,
 	ctx->query_atom = query;
 
 	/* handle Packages (binpkgs index) file separately */
-	if (ctx->cachetype == CACHE_PACKAGES)
+	if (ctx->treetype == TREE_PACKAGES)
 		return tree_foreach_packages(ctx, callback, priv);
 
+#ifdef ENABLE_GTREE
+	/* similar a gtree cache can be read sequentially in one go */
+	if (ctx->treetype == TREE_METADATA_GTREE)
+		return tree_foreach_gtree(ctx, callback, priv);
+	if (ctx->treetype == TREE_EBUILD &&
+		ctx->subtree != NULL &&
+		ctx->subtree->treetype == TREE_METADATA_GTREE)
+		return tree_foreach_gtree(ctx->subtree, callback, priv);
+#endif
+
 	ret = 0;
 	while ((cat_ctx = tree_next_cat(ctx))) {
 		while ((pkg_ctx = tree_next_pkg(cat_ctx))) {
@@ -1603,7 +2098,7 @@ tree_get_atom(tree_pkg_ctx *pkg_ctx, bool complete)
 
 	if (complete) {
 		tree_ctx *ctx = pkg_ctx->cat_ctx->ctx;
-		if (ctx->cachetype == CACHE_VDB) {
+		if (ctx->treetype == TREE_VDB) {
 			if (pkg_ctx->atom->SLOT == NULL) {
 				/* FIXME: use tree_meta_get !!! */
 				if (pkg_ctx->slot == NULL)
@@ -1633,7 +2128,7 @@ tree_get_atom(tree_pkg_ctx *pkg_ctx, bool complete)
 			}
 			/* repo is set from the tree, when found */
 			if (pkg_ctx->atom->REPO == NULL) {
-				if (pkg_ctx->repo == NULL && ctx->cachetype == CACHE_BINPKGS) {
+				if (pkg_ctx->repo == NULL && ctx->treetype == TREE_BINPKGS) {
 					if (meta == NULL)
 						meta = tree_pkg_read(pkg_ctx);
 					if (meta != NULL && meta->Q_repository != NULL) {
@@ -1722,18 +2217,20 @@ tree_match_atom_cache_populate_cb(tree_pkg_ctx *ctx, void *priv)
 		cat_ctx->name = contains_set(atom->CATEGORY, cache);
 	}
 
+	pkg = xcalloc(1, sizeof(*pkg));
+
 	/* FIXME: this really could use a set */
 	cat_ctx->pkg_cnt++;
 	cat_ctx->pkg_ctxs = xrealloc(cat_ctx->pkg_ctxs,
 			sizeof(*cat_ctx->pkg_ctxs) * cat_ctx->pkg_cnt);
-	cat_ctx->pkg_ctxs[cat_ctx->pkg_cnt - 1] =
-		pkg = tree_open_pkg(cat_ctx, atom->PF);
+	cat_ctx->pkg_ctxs[cat_ctx->pkg_cnt - 1] = pkg;
+	pkg->cat_ctx = cat_ctx;
 	pkg->atom = atom_clone(atom);
 	pkg->name = xstrdup(pkg->atom->PF);
 	pkg->repo = tctx->repo != NULL ? xstrdup(tctx->repo) : NULL;
 	if (meta != NULL) {
 		pkg->fd = -2;  /* don't try to read, we fill it in here */
-		if (tctx->cachetype == CACHE_PACKAGES) {
+		if (tctx->treetype == TREE_PACKAGES) {
 			/* need to copy, source is based on temp space in foreach */
 			pkg->meta = tree_clone_meta(meta);
 		} else {
@@ -1741,6 +2238,7 @@ tree_match_atom_cache_populate_cb(tree_pkg_ctx *ctx, void *priv)
 			 * xpak archive, so can just take it over */
 			pkg->meta = meta;
 			ctx->meta = NULL;  /* avoid double free */
+			pkg->binpkg_isgpkg = ctx->binpkg_isgpkg;
 		}
 	} else {
 		pkg->meta = NULL;
@@ -1749,19 +2247,86 @@ tree_match_atom_cache_populate_cb(tree_pkg_ctx *ctx, void *priv)
 	return 0;
 }
 
+static tree_match_ctx *
+tree_match_search_cat_int(
+		tree_cat_ctx      *cat_ctx,
+		const depend_atom *query,
+		int                flags
+)
+{
+	tree_pkg_ctx   *pkg_ctx;
+	tree_match_ctx *ret      = NULL;
+	depend_atom    *atom;
+	char           *lastpn   = NULL;
+
+	while ((pkg_ctx = tree_next_pkg(cat_ctx)) != NULL) {
+		atom = tree_get_atom(pkg_ctx,
+							 (query->SLOT != NULL ||
+							  flags & TREE_MATCH_FULL_ATOM));
+		/* skip virtual/ package as requested */
+		if (!(flags & TREE_MATCH_VIRTUAL ||
+			  strcmp(atom->CATEGORY, "virtual") != 0))
+			continue;
+		/* skip acct-* package as requested */
+		if (!(flags & TREE_MATCH_ACCT ||
+			  strncmp(atom->CATEGORY, "acct-", sizeof("acct-") - 1) != 0))
+			continue;
+		/* see if this atom matches the query */
+		if (atom_compare(atom, query) == EQUAL) {
+			tree_match_ctx *n;
+			/* skip over additional versions for match latest */
+			if (flags & TREE_MATCH_LATEST && lastpn != NULL &&
+				strcmp(lastpn, atom->PN) == 0)
+				continue;
+			/* create a new match result */
+			n = xzalloc(sizeof(tree_match_ctx));
+			n->atom = atom;
+			n->pkg = pkg_ctx;
+			if (cat_ctx->ctx->treetype == TREE_PACKAGES &&
+				pkg_ctx->meta->Q_PATH != NULL)
+			{
+				/* binpkg-multi-instance has a PATH ready for us */
+				snprintf(n->path, sizeof(n->path), "%s/%s",
+						 (char *)cat_ctx->ctx->path, pkg_ctx->meta->Q_PATH);
+			} else {
+				snprintf(n->path, sizeof(n->path), "%s/%s/%s%s",
+						 (char *)cat_ctx->ctx->path,
+						 cat_ctx->name, pkg_ctx->name,
+						 cat_ctx->ctx->treetype == TREE_EBUILD   ? ".ebuild" :
+						 (cat_ctx->ctx->treetype == TREE_BINPKGS ||
+						  cat_ctx->ctx->treetype == TREE_PACKAGES) ?
+						 (pkg_ctx->binpkg_isgpkg   ? ".gpkg.tar" : ".tbz2")  :
+						                                           "");
+			}
+			if (flags & TREE_MATCH_METADATA)
+				n->meta = tree_pkg_read(pkg_ctx);
+			if (cat_ctx->ctx->treetype == TREE_BINPKGS ||
+				cat_ctx->ctx->treetype == TREE_PACKAGES)
+				n->free_atom = n->free_meta = 0;
+			n->next = ret;
+			ret = n;
+			lastpn = atom->PN;
+		}
+		if (flags & TREE_MATCH_FIRST && ret != NULL)
+			break;
+	}
+	cat_ctx->pkg_cur = 0;  /* reset to allow another traversal */
+
+	return ret;
+}
+
 tree_match_ctx *
 tree_match_atom(tree_ctx *ctx, const depend_atom *query, int flags)
 {
 	tree_cat_ctx *cat_ctx;
-	tree_pkg_ctx *pkg_ctx;
 	tree_match_ctx *ret = NULL;
-	depend_atom *atom;
 
 	ctx->do_sort = true;     /* sort uses buffer, which cache relies on */
 	ctx->query_atom = NULL;  /* ensure the cache contains ALL pkgs */
 
-	if ((ctx->cachetype == CACHE_PACKAGES || ctx->cachetype == CACHE_BINPKGS)
-			&& ctx->cache.categories == NULL)
+	if ((ctx->treetype == TREE_PACKAGES ||
+		 ctx->treetype == TREE_BINPKGS) &&
+		ctx->cache.categories == NULL)
 	{
 		set *cache;
 		DECLARE_ARRAY(cats);
@@ -1783,9 +2348,8 @@ tree_match_atom(tree_ctx *ctx, const depend_atom *query, int flags)
 		tree_foreach_pkg(ctx,
 				tree_match_atom_cache_populate_cb, cache, true, NULL);
 
-		ctx->do_sort = true;  /* turn it back on */
+		ctx->do_sort = true;  /* turn it back on after tree_foreach_pkg */
 		ctx->cache.all_categories = true;
-
 		ctx->cache.categories = cache;
 
 		/* loop through all categories, and sort the pkgs */
@@ -1805,63 +2369,13 @@ tree_match_atom(tree_ctx *ctx, const depend_atom *query, int flags)
 	if (ctx->cache.categories == NULL)
 		ctx->cache.categories = create_set();
 
-#define search_cat(C) \
-{ \
-	char *lastpn = NULL; \
-	while ((pkg_ctx = tree_next_pkg(C)) != NULL) { \
-		atom = tree_get_atom(pkg_ctx, \
-				query->SLOT != NULL || flags & TREE_MATCH_FULL_ATOM); \
-		/* skip virtual/ package as requested */ \
-		if (!(flags & TREE_MATCH_VIRTUAL || \
-				strcmp(atom->CATEGORY, "virtual") != 0)) \
-			continue; \
-		/* skip acct-* package as requested */ \
-		if (!(flags & TREE_MATCH_ACCT || \
-				strncmp(atom->CATEGORY, "acct-", sizeof("acct-") - 1) != 0)) \
-			continue; \
-		/* see if this atom matches the query */ \
-		if (atom_compare(atom, query) == EQUAL) { \
-			tree_match_ctx *n; \
-			/* skip over additional versions for match latest */ \
-			if (flags & TREE_MATCH_LATEST && lastpn != NULL && \
-					strcmp(lastpn, atom->PN) == 0) \
-				continue; \
-			/* create a new match result */ \
-			n = xzalloc(sizeof(tree_match_ctx)); \
-			n->atom = atom; \
-			n->pkg = pkg_ctx; \
-			if (C->ctx->cachetype == CACHE_PACKAGES && \
-				pkg_ctx->meta->Q_PATH != NULL) \
-			{ \
-				/* binpkg-multi-instance has a PATH ready for us */ \
-				snprintf(n->path, sizeof(n->path), "%s/%s", \
-						 (char *)C->ctx->path, pkg_ctx->meta->Q_PATH); \
-			} else { \
-				snprintf(n->path, sizeof(n->path), "%s/%s/%s%s", \
-						 (char *)C->ctx->path, C->name, pkg_ctx->name, \
-						 C->ctx->cachetype == CACHE_EBUILD   ? ".ebuild" : \
-						 C->ctx->cachetype == CACHE_BINPKGS  ? ".tbz2" : \
-						 C->ctx->cachetype == CACHE_PACKAGES ? ".tbz2" : ""); \
-			} \
-			if (flags & TREE_MATCH_METADATA) \
-				n->meta = tree_pkg_read(pkg_ctx); \
-			if (C->ctx->cachetype == CACHE_BINPKGS || \
-					C->ctx->cachetype == CACHE_PACKAGES) \
-				n->free_atom = n->free_meta = 0; \
-			n->next = ret; \
-			ret = n; \
-			lastpn = atom->PN; \
-		} \
-		if (flags & TREE_MATCH_FIRST && ret != NULL) \
-			break; \
-	} \
-	C->pkg_cur = 0;  /* reset to allow another traversal */ \
-}
-
 	if (query->CATEGORY == NULL) {
+		tree_match_ctx *tret;
 		/* loop through all cats */
 		while ((cat_ctx = tree_next_cat(ctx)) != NULL) {
-			search_cat(cat_ctx);
+			tret = tree_match_search_cat_int(cat_ctx, query, flags);
+			if (tret != NULL)
+				ret = tret;
 			if (ret != NULL && flags & TREE_MATCH_FIRST)
 				break;
 		}
@@ -1870,7 +2384,7 @@ tree_match_atom(tree_ctx *ctx, const depend_atom *query, int flags)
 	} else {
 		/* try CAT, and PN for latest version */
 		if ((cat_ctx = tree_open_cat(ctx, query->CATEGORY)) != NULL)
-			search_cat(cat_ctx);
+			ret = tree_match_search_cat_int(cat_ctx, query, flags);
 	}
 
 	return ret;
