@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2023 Gentoo Foundation
+ * Copyright 2005-2025 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  */
 
@@ -31,14 +31,16 @@ struct tree_ctx {
 	size_t cat_cur;
 	bool do_sort:1;
 	enum {
-		CACHE_UNSET = 0,
-		CACHE_METADATA_MD5,
-		CACHE_METADATA_PMS,
-		CACHE_EBUILD,
-		CACHE_VDB,
-		CACHE_PACKAGES,
-		CACHE_BINPKGS,
-	} cachetype:3;
+		TREE_UNSET = 0,
+		TREE_METADATA_GTREE,
+		TREE_METADATA_MD5,
+		TREE_METADATA_PMS,
+		TREE_EBUILD,
+		TREE_VDB,
+		TREE_PACKAGES,
+		TREE_BINPKGS,
+	} treetype:3;
+	tree_ctx *subtree;
 	tree_pkg_ctx *ebuilddir_pkg_ctx;
 	tree_cat_ctx *ebuilddir_cat_ctx;
 	tree_ctx *ebuilddir_ctx;
@@ -73,6 +75,7 @@ struct tree_pkg_ctx {
 	size_t slot_len;
 	size_t repo_len;
 	int fd;
+	int binpkg_isgpkg:1;
 	tree_cat_ctx *cat_ctx;
 	depend_atom *atom;
 	tree_pkg_meta *meta;
@@ -152,6 +155,8 @@ tree_ctx *tree_open(const char *sroot, const char *portdir);
 tree_ctx *tree_open_vdb(const char *sroot, const char *svdb);
 tree_ctx *tree_open_ebuild(const char *sroot, const char *portdir);
 tree_ctx *tree_open_binpkg(const char *sroot, const char *spkg);
+#define tree_open_gtree(R,T) tree_open_gtree_int(R,T,false)
+tree_ctx *tree_open_gtree_int(const char *sroot, const char *tar, bool quiet);
 void tree_close(tree_ctx *ctx);
 tree_cat_ctx *tree_open_cat(tree_ctx *ctx, const char *name);
 void tree_close_cat(tree_cat_ctx *cat_ctx);
