@@ -1,5 +1,25 @@
+/*
+ * Copyright 2011-2019 Gentoo Foundation
+ * Distributed under the terms of the GNU General Public License v2
+ *
+ * Copyright 2011-2016 Mike Frysinger  - <vapier@gentoo.org>
+ * Copyright 2017-     Fabian Groffen  - <grobian@gentoo.org>
+ */
+
+#include "main.h"
+
+#include <stdlib.h>
+#include <unistd.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <dirent.h>
+#include <fcntl.h>
+#include <xalloc.h>
+
+#include "xmkdir.h"
+
 /* Emulate `mkdir -p -m MODE PATH` */
-static int
+int
 mkdir_p_at(int dfd, const char *path, mode_t mode)
 {
 	char *_p, *p, *s;
@@ -36,14 +56,15 @@ mkdir_p_at(int dfd, const char *path, mode_t mode)
 
 	return 0;
 }
-static int
+
+int
 mkdir_p(const char *path, mode_t mode)
 {
 	return mkdir_p_at(AT_FDCWD, path, mode);
 }
 
 /* Emulate `rm -rf PATH` */
-static int
+int
 rm_rf_at(int dfd, const char *path)
 {
 	int subdfd;
@@ -86,19 +107,20 @@ rm_rf_at(int dfd, const char *path)
 	return ret;
 }
 
-static int
+int
 rm_rf(const char *path)
 {
 	return rm_rf_at(AT_FDCWD, path);
 }
 
-static int
+int
 rmdir_r_at(int dfd, const char *path)
 {
 	size_t len;
 	char *p, *e;
 
-	p = xstrdup_len(path, &len);
+	p = xstrdup(path);
+	len = strlen(p);
 	e = p + len;
 
 	while (e != p) {
@@ -114,10 +136,8 @@ rmdir_r_at(int dfd, const char *path)
 	return 0;
 }
 
-/*
-static int
+int
 rmdir_r(const char *path)
 {
 	return rmdir_r_at(AT_FDCWD, path);
 }
-*/
