@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2023 Gentoo Foundation
+ * Copyright 2005-2025 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2008 Ned Ludd        - <solar@gentoo.org>
@@ -72,7 +72,7 @@ init_coredumps(void)
 }
 #endif
 
-void
+static void
 setup_quiet(void)
 {
 	/* "e" for FD_CLOEXEC */
@@ -106,13 +106,13 @@ usage(int status, const char *flags, struct option const opts[],
 				fprintf(fp, " %s%9s%s %s%-16s%s%s:%s %s\n",
 					YELLOW, applets[i].name, NORM,
 					DKBLUE, applets[i].opts, NORM,
-					RED, NORM, _(applets[i].desc));
+					RED, NORM, applets[i].desc);
 	} else if (blabber > 0) {
 		fprintf(fp, "%susage:%s %s%s%s [opts] %s%s%s %s:%s %s\n",
 			GREEN, NORM,
 			YELLOW, applets[blabber].name, NORM,
 			DKBLUE, applets[blabber].opts, NORM,
-			RED, NORM, _(applets[blabber].desc));
+			RED, NORM, applets[blabber].desc);
 		if (desc)
 			fprintf(fp, "\n%s\n", desc);
 	}
@@ -154,10 +154,10 @@ usage(int status, const char *flags, struct option const opts[],
 		/* then wrap the help text, if necessary */
 		prefixlen = 6 + 2 + optlen + 1 + 1 + 1;
 		if ((size_t)twidth < prefixlen + 10) {
-			fprintf(fp, "%s\n", _(help[i]));
+			fprintf(fp, "%s\n", help[i]);
 		} else {
 			const char *t;
-			hstr = _(help[i]);
+			hstr = help[i];
 			l = strlen(hstr);
 			while (twidth - prefixlen < l) {
 				/* search backwards for a space */
@@ -428,7 +428,7 @@ static void
 read_portage_file(const char *file, enum portage_file_type type, void *data)
 {
 	FILE *fp;
-	struct dirent **dents;
+	struct dirent **dents = NULL;
 	int dentslen;
 	char *s;
 	char *p;
@@ -443,7 +443,8 @@ read_portage_file(const char *file, enum portage_file_type type, void *data)
 	env_vars *vars = data;
 	set *masks = data;
 
-	if ((dentslen = scandir(file, &dents, NULL, alphasort)) > 0) {
+	snprintf(npath, sizeof(npath), "%s%s", portroot, file);
+	if ((dentslen = scandir(npath, &dents, NULL, alphasort)) > 0) {
 		int di;
 		struct dirent *d;
 
@@ -460,7 +461,7 @@ read_portage_file(const char *file, enum portage_file_type type, void *data)
 		goto done;
 	}
 
-	fp = fopen(file, "r");
+	fp = fopen(npath, "r");
 	if (fp == NULL)
 		goto done;
 
@@ -603,7 +604,7 @@ read_portage_file(const char *file, enum portage_file_type type, void *data)
 	free(buf);
 
 	if (getenv("DEBUG"))
-		fprintf(stderr, "read profile %s\n", file);
+		fprintf(stderr, "read profile %s%s\n", portroot, file);
 }
 
 /* Helper to check if a string starts with a prefix. If so, returns
@@ -756,9 +757,9 @@ read_portage_profile(const char *profile, env_vars vars[], set *masks)
 
 	/* now consume *this* profile's make.defaults and package.mask */
 	strcpy(profile_file + profile_len, "make.defaults");
-	read_portage_file(profile_file, ENV_FILE, vars);
+	read_portage_file(profile_file + 1, ENV_FILE, vars);
 	strcpy(profile_file + profile_len, "package.mask");
-	read_portage_file(profile_file, PMASK_FILE, masks);
+	read_portage_file(profile_file + 1, PMASK_FILE, masks);
 }
 
 env_vars vars_to_read[] = {
@@ -786,7 +787,7 @@ env_vars vars_to_read[] = {
 	_Q_EVS(STR,  ARCH,                portarch,            "")
 	_Q_EVS(ISTR, CONFIG_PROTECT,      config_protect,      "/etc")
 	_Q_EVS(ISTR, CONFIG_PROTECT_MASK, config_protect_mask, "")
-	_Q_EVB(BOOL, NOCOLOR,             nocolor,             0)
+	_Q_EVB(BOOL, NOCOLOR,             nocolor,             false)
 	_Q_EVT(ISET, FEATURES,            features,            NULL)
 	_Q_EVT(ISET, USE,                 ev_use,              NULL)
 	_Q_EVS(STR,  EPREFIX,             eprefix,             CONFIG_EPREFIX)
@@ -883,10 +884,18 @@ read_one_repos_conf(const char *repos_conf, char **primary)
 			void *ele;
 			size_t n;
 			char *overlay;
+			char rootovrl[_Q_PATH_MAX];
 
 			/* try not to get confused by symlinks etc. */
-			if (realpath(e, rrepo) != NULL)
-				e = rrepo;
+			if (*e == '/')
+				e++;  /* ROOT ends with / */
+			snprintf(rootovrl, sizeof(rootovrl), "%s%s", portroot, e);
+			n = strlen(portroot);
+			if (realpath(rootovrl, rrepo) != NULL &&
+				strncmp(rrepo, portroot, n) == 0)
+				e = rrepo + n - 1;
+			else
+				e = rootovrl + n - 1;
 
 			array_for_each(overlay_names, n, overlay) {
 				if (strcmp(overlay, repo) == 0)
@@ -922,7 +931,7 @@ read_repos_conf(const char *configroot, const char *repos_conf, char **primary)
 	int i, count;
 	struct dirent **confs;
 
-	xasprintf(&top_conf, "%s%s", configroot, repos_conf);
+	xasprintf(&top_conf, "%s%s%s", portroot, configroot, repos_conf);
 	if (getenv("DEBUG"))
 		fprintf(stderr, "repos.conf.d scanner %s\n", top_conf);
 	count = scandir(top_conf, &confs, NULL, alphasort);
@@ -969,24 +978,16 @@ initialize_portage_env(void)
 	const char *configroot = getenv("PORTAGE_CONFIGROOT");
 	char *primary_overlay = NULL;
 
-	/* initialize all the properties with their default value */
-	for (i = 0; vars_to_read[i].name; ++i) {
-		var = &vars_to_read[i];
-		switch (var->type) {
-			case _Q_BOOL:  *var->value.b = var->default_value;           break;
-			case _Q_STR:
-			case _Q_ISTR:  *var->value.s = xstrdup(var->default_value);  break;
-			case _Q_ISET:  *var->value.t = (set *)var->default_value;    break;
-		}
-		var->src = xstrdup(STR_DEFAULT);
-	}
-
 	package_masks = create_set();
 
 	/* figure out where to find our config files */
 	if (!configroot)
 		configroot = CONFIG_EPREFIX;
 
+	if (*configroot != '/')
+		err("PORTAGE_CONFIGROOT must be an absolute path");
+	configroot++;  /* strip leading /, ROOT always ends with one */
+
 	/* rstrip /-es */
 	i = strlen(configroot);
 	while (i > 0 && configroot[i - 1] == '/')
@@ -1013,20 +1014,20 @@ initialize_portage_env(void)
 			if (overlay == primary_overlay) {
 				snprintf(pathbuf, sizeof(pathbuf), "%s/profiles/package.mask",
 						(char *)array_get_elem(overlays, n));
-				read_portage_file(pathbuf, PMASK_FILE, package_masks);
+				read_portage_file(pathbuf + 1, PMASK_FILE, package_masks);
 				break;
 			}
 		}
 	}
 
 	/* walk all the stacked profiles */
-	snprintf(pathbuf, sizeof(pathbuf), "%.*s/etc/make.profile",
-			(int)i, configroot);
+	snprintf(pathbuf, sizeof(pathbuf), "%s%.*s/etc/make.profile",
+			 portroot, (int)i, configroot);
 	read_portage_profile(
 			realpath(pathbuf, rpathbuf) == NULL ? pathbuf : rpathbuf,
 			vars_to_read, package_masks);
-	snprintf(pathbuf, sizeof(pathbuf), "%.*s/etc/portage/make.profile",
-			(int)i, configroot);
+	snprintf(pathbuf, sizeof(pathbuf), "%s%.*s/etc/portage/make.profile",
+			 portroot, (int)i, configroot);
 	read_portage_profile(
 			realpath(pathbuf, rpathbuf) == NULL ? pathbuf : rpathbuf,
 			vars_to_read, package_masks);
@@ -1128,13 +1129,16 @@ initialize_portage_env(void)
 	 * administration in overlays */
 	{
 		char *overlay;
-		var = &vars_to_read[11];  /* PORTDIR */
+		var = &vars_to_read[12];  /* PORTDIR */
 
 		if (strcmp(var->src, STR_DEFAULT) != 0 || array_cnt(overlays) == 0) {
 			char roverlay[_Q_PATH_MAX];
+			char rootovrl[_Q_PATH_MAX];
+			snprintf(rootovrl, sizeof(rootovrl), "%s%s",
+					 portroot, main_overlay);
 			/* get cannonical path, we do so for repos.conf too */
-			if (realpath(main_overlay, roverlay) == NULL)
-				snprintf(roverlay, sizeof(roverlay), "%s", main_overlay);
+			if (realpath(rootovrl, roverlay) == NULL)
+				snprintf(roverlay, sizeof(roverlay), "%s", rootovrl);
 			array_for_each(overlays, i, overlay) {
 				if (strcmp(overlay, roverlay) == 0)
 					break;
@@ -1230,53 +1234,83 @@ int main(int argc, char **argv)
 	IF_DEBUG(init_coredumps());
 	argv0 = argv[0];
 
-	setlocale(LC_ALL, "");
-	bindtextdomain(argv0, CONFIG_EPREFIX "usr/share/locale");
-	textdomain(argv0);
+	/* ensure any err/warn doesn't use unitialised vars */
+	color_clear();
 
-	/* note: setting nocolor here is pointless, since
-	 * initialize_portage_env is going to re-init nocolor, so make
-	 * sure we modify the default instead. */
+	/* initialise all the properties with their default value */
+	for (i = 0; vars_to_read[i].name; ++i) {
+		env_vars *var = &vars_to_read[i];
+		switch (var->type) {
+			case _Q_BOOL:  *var->value.b = var->default_value;           break;
+			case _Q_STR:
+			case _Q_ISTR:  *var->value.s = xstrdup(var->default_value);  break;
+			case _Q_ISET:  *var->value.t = (set *)var->default_value;    break;
+		}
+		var->src = xstrdup(STR_DEFAULT);
+	}
+
+	/* disable colours when not writing to a terminal */
 	twidth = 0;
-	nocolor = 0;
+	nocolor = false;
 	if (fstat(fileno(stdout), &st) != -1) {
 		if (!isatty(fileno(stdout))) {
-			nocolor = 1;
+			nocolor = true;
 		} else {
 			if ((getenv("TERM") == NULL) ||
 					(strcmp(getenv("TERM"), "dumb") == 0))
-				nocolor = 1;
+				nocolor = true;
 			if (ioctl(0, TIOCGWINSZ, &winsz) == 0 && winsz.ws_col > 0)
 				twidth = (int)winsz.ws_col;
 		}
 	} else {
-		nocolor = 1;
+		nocolor = true;
 	}
-	vars_to_read[7].default_value = (char *)nocolor;  /* NOCOLOR */
+	if (nocolor)
+		set_portage_env_var(&vars_to_read[7], "true", "terminal"); /* NOCOLOR */
 
 	/* We can use getopt here, but only in POSIX mode (which stops at
 	 * the first non-option argument) because otherwise argv is
 	 * modified, this basically sulks, because ideally we parse and
 	 * handle the common options here.  Because we are parsing profiles
-	 * and stuff at this point we need -q for bug #735134, so do lame
-	 * matching for that */
+	 * and stuff at this point we need -q for bug #735134, and --root so
+	 * do lame matching for that */
 	for (i = 1; i < argc; i++) {
-		if (argv[i] != NULL && argv[i][0] == '-' &&
-			(argv[i][1] == 'q' || strcmp(&argv[i][1], "-quiet") == 0))
-		{
-			setup_quiet();
+		if (argv[i] != NULL && argv[i][0] == '-') {
+			if (argv[i][1] == '-') {
+				if (strcmp(&argv[i][2], "quiet") == 0) {
+					setup_quiet();
+				} else if (strcmp(&argv[i][2], "root") == 0 &&
+						 argv[i + 1] != NULL)
+				{
+					char  realroot[_Q_PATH_MAX];
+					char *root;
+					if (realpath(argv[i + 1], realroot) != NULL)
+						root = realroot;
+					else
+						errp("--root argument could not be resolved");
+					set_portage_env_var(&vars_to_read[0], root,
+										"command line");  /* ROOT */
+				}
+			} else {
+				char *p;
+
+				/* must handle combinations of options like -Rq :( */
+				for (p = &argv[i][1]; *p != '\0'; p++) {
+					switch (*p) {
+						case 'q':
+							setup_quiet();
+							break;
+					}
+				}
+			}
 		}
 	}
+
 	/* same for env-based fallback */
 	if (getenv("PORTAGE_QUIET") != NULL)
 		setup_quiet();
 
-	/* ensure color strings are initialised, early code here may use
-	 * e.g. warn which uses them */
-	color_clear();
-
 	initialize_portage_env();
 	optind = 0;
-	quiet  = 0;
 	return q_main(argc, argv);
 }
