@@ -295,6 +295,8 @@ qdepends_results_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 			}
 		}
 
+		array_for_each(state->deps, m, atom)
+			atom_implode(atom);
 		xarrayfree_int(state->deps);
 		dep_burn_tree(dep_tree);
 	}
@@ -416,6 +418,7 @@ int qdepends_main(int argc, char **argv)
 		array_for_each(overlays, n, overlay) {
 			t = tree_open(portroot, overlay);
 			if (t != NULL) {
+				state.rtree = NULL;
 				if (state.resolve)
 					state.rtree = tree_open(portroot, overlay);
 				if (!(state.qmode & QMODE_REVERSE) && array_cnt(atoms) > 0) {
