@@ -1,5 +1,5 @@
 /*
- * Copyright 2018-2024 Gentoo Foundation
+ * Copyright 2018-2025 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2018-     Fabian Groffen  - <grobian@gentoo.org>
@@ -29,8 +29,6 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/time.h>
-#include <openssl/sha.h>
-#include <blake2.h>
 #include <zlib.h>
 #include <gpgme.h>
 
@@ -185,9 +183,9 @@ write_hashes(
 		gzFile gm)
 {
 	size_t flen = 0;
-	char sha256[(SHA256_DIGEST_LENGTH * 2) + 1];
-	char sha512[(SHA512_DIGEST_LENGTH * 2) + 1];
-	char blak2b[(BLAKE2B_OUTBYTES * 2) + 1];
+	char sha256[SHA256_DIGEST_LENGTH + 1];
+	char sha512[SHA512_DIGEST_LENGTH + 1];
+	char blak2b[BLAKE2B_DIGEST_LENGTH + 1];
 	char data[8192];
 	char fname[8192];
 	size_t len;
@@ -929,7 +927,9 @@ verify_gpg_sig(const char *path, verify_msg **msgs)
 
 	/* we only check/return the first signature */
 	if ((sig = vres->signatures) != NULL) {
-		ret = xmalloc(sizeof(gpg_sig));
+		bool cleanup = false;
+
+		ret = xcalloc(1, sizeof(gpg_sig));
 
 		if (sig->fpr != NULL) {
 			snprintf(buf, sizeof(buf),
@@ -984,24 +984,31 @@ verify_gpg_sig(const char *path, verify_msg **msgs)
 						"used to verify the signature has been revoked");
 				break;
 			case GPG_ERR_BAD_SIGNATURE:
-				free(ret);
-				ret = NULL;
+				cleanup = true;
 				printf("the signature is invalid\n");
 				break;
 			case GPG_ERR_NO_PUBKEY:
-				free(ret);
-				ret = NULL;
+				cleanup = true;
 				printf("the signature could not be verified due to a "
 						"missing key for:\n  %s\n", buf);
 				break;
 			default:
-				free(ret);
-				ret = NULL;
+				cleanup = true;
 				printf("there was some error which prevented the "
 						"signature verification:\n  %s: %s\n",
 						buf, gpgme_strerror(sig->status));
 				break;
 		}
+
+		if (cleanup) {
+			free(ret->algo);
+			free(ret->fingerprint);
+			free(ret->timestamp);
+			free(ret->signer);
+			free(ret->pkfingerprint);
+			free(ret);
+			ret = NULL;
+		}
 	}
 
 	gpgme_data_release(out);
@@ -1027,9 +1034,9 @@ verify_file(const char *dir, char *mfline, const char *mfest, verify_msg **msgs)
 	char *p;
 	char buf[8192];
 	size_t flen = 0;
-	char sha256[(SHA256_DIGEST_LENGTH * 2) + 1];
-	char sha512[(SHA512_DIGEST_LENGTH * 2) + 1];
-	char blak2b[(BLAKE2B_OUTBYTES * 2) + 1];
+	char sha256[SHA256_DIGEST_LENGTH + 1];
+	char sha512[SHA512_DIGEST_LENGTH + 1];
+	char blak2b[BLAKE2B_DIGEST_LENGTH + 1];
 	char ret = 0;
 
 	/* mfline is a Manifest file line with type and leading path
@@ -1357,8 +1364,8 @@ verify_dir(
 			}
 		}
 
-		while (dentrieslen-- > 0)
-			free(dentries[dentrieslen]);
+		for (curdentry = 0; curdentry < dentrieslen; curdentry++)
+			free(dentries[curdentry]);
 		free(dentries);
 
 #pragma omp parallel for shared(ret) private(entry, etpe, slash)
@@ -1497,8 +1504,8 @@ verify_manifest(
 	ret = verify_dir(dir, elems, elemslen, 0, buf + 2, msgs);
 	checked_manifests++;
 
-	while (elemslen-- > 0)
-		free(elems[elemslen]);
+	for (elemssize = 0; elemssize < elemslen; elemssize++)
+		free(elems[elemssize]);
 	free(elems);
 
 	return ret;
