@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2024 Gentoo Foundation
+ * Copyright 2005-2025 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2010 Ned Ludd        - <solar@gentoo.org>
@@ -56,12 +56,6 @@ struct qtegrity_opt_state {
 #define FILE_EMPTY 2
 #define FILE_RELATIVE 3
 
-#define SHA1_DIGEST_LENGTH 40
-#define SHA256_PREFIX_LENGTH 8
-#define SHA256_DIGEST_LENGTH 64
-#define SHA256_LENGTH (SHA256_PREFIX_LENGTH + SHA256_DIGEST_LENGTH)
-#define SHA512_DIGEST_LENGTH 128
-
 static void
 check_sha(char *ret_digest, char *path, char *algo)
 {
@@ -247,6 +241,7 @@ int qtegrity_main(int argc, char **argv)
 				state.ima = false;
 				state.add = true;
 				if (check_file(optarg) == FILE_SUCCESS) {
+					free(state.add_file);
 					state.add_file = xstrdup(optarg);
 				} else {
 					err("Expected absolute file as argument, got '%s'", optarg);
